﻿using Kivii.Finances.Entities;
using Kivii.Linq;
using Kivii.Web;
using System;
using System.Collections.Generic;
using System.Data;
using System.Linq;
using System.Linq.Expressions;
using System.Text;
using System.Threading.Tasks;

namespace Kivii.Finances.Transforms
{
    [Api(Description = "创建账单")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class BillCreate : RestfulExecution<Bill>
    {
        public List<Guid> SettlementKvids { get; set; }

        public Bill Item { get; set; }

        public override object OnExecution(IRequest req, IResponse res)
        {
            if (Item.PayerName.IsNullOrEmpty() || Item.PayeeName.IsNullOrEmpty()) throw new Exception("收付双方不能为空!");
            if (Item.Currency == CurrencyUnit.Unsupported) throw new Exception("未传入货币单位!");
            if (SettlementKvids.IsNullOrEmpty()) throw new Exception("未传入结算数据");
            var conn = KiviiContext.GetOpenedDbConnection<Bill>();
            conn.InitEntityType<Bill>();
            conn.InitEntityType<BillDetail>();

            var query = conn.From<Settlement>();
            query.Where(o => Sql.In(o.Kvid, SettlementKvids));
            query.Select(o => new
            {
                o.SerialNumber,
                o.Kvid,
                o.Amount,
                o.Currency,
                o.Type,
                o.Summary,
                o.Remark
            });
            var settlements = conn.Select(query);
            if (settlements == null) throw new Exception("无对应数据导出!");
            (settlements.Exists(o => o.Currency != Item.Currency)).ThrowIfTrue("存在于账单不一致的货币单位不可制作!");
            if (Item.Amount != settlements.Sum(o => o.Amount)) throw new Exception("账单金额不等于选择的结算金额之和!");
            if (conn.Exists<BillDetail>(o => o.OffsetKvid == Guid.Empty && Sql.In(o.BizKvid, settlements.ConvertAll(p => p.Kvid)))) throw new Exception("此结算已有账单,不可重复制作账单!");
            var rtns = new RestfulCreateResponse<Bill>();
            rtns.Results = new List<Bill>();

            var trans = conn.OpenTransaction();
            try
            {
                var bill = Item.Accept(settlements, conn);
                bill.RemoveAllOnlyProperties();
                rtns.Results.Add(bill);

                trans.Commit();
            }
            catch (Exception ex)
            {
                trans.Rollback();
                throw ex;
            }

            return rtns;
        }
    }

    [RequiresAnyRole(SystemRoles.Everyone)]
    public class BillUpdate : RestfulUpdate<Bill>
    {
    }

    [RequiresAnyRole(SystemRoles.Everyone)]
    public class BillRead : RestfulRead<Bill>
    {
        public bool IncludeDetail { get; set; }
        public override bool OnPostRestfulRead(IRequest req, IResponse res, IDbConnection dbConnection, IRestfulReadResponse<Bill> rtns)
        {
            if (!IncludeDetail) return base.OnPostRestfulRead(req, res, dbConnection, rtns);
            var queryDetails = dbConnection.From<BillDetail>();
            queryDetails.Where(o => o.BillKvid == rtns.Result.Kvid);
            queryDetails.Select(o => o.BizKvid);
            rtns.Result.Details = dbConnection.Select<Settlement>(o => Sql.In(o.Kvid, queryDetails));
            return base.OnPostRestfulRead(req, res, dbConnection, rtns);
        }
    }

    [RequiresAnyRole(SystemRoles.Everyone)]
    public class BillQuery : RestfulQuery<Bill>
    {
        /// <summary>
        /// true:只查已经付款的账单
        /// false:只查未付款的账单
        /// 不传入: 全部账单
        /// </summary>
        public bool? IsPayed { get; set; }

        public override bool OnPreRestfulQuery(IRequest req, IResponse res, IDbConnection dbConnection, IRestfulQueryResponse<Bill> rtns, ref Expression<Func<Bill, bool>> where)
        {
            if (IsPayed == null) return base.OnPreRestfulQuery(req, res, dbConnection, rtns, ref where);
            else if (IsPayed.Value) where = o => o.Amount == o.AmountPayment;
            else where = o => o.Amount > o.AmountPayment || o.AmountPayment == 0;
            return base.OnPreRestfulQuery(req, res, dbConnection, rtns, ref where);
        }
    }

    [RequiresAnyRole(SystemRoles.Everyone)]
    public class BillDelete : RestfulDelete<Bill>
    {
        IDbTransaction _trans = null;
        public override bool OnPreRestfulDelete(IRequest req, IResponse res, IDbConnection dbConnection, IRestfulDeleteResponse<Bill> rtns)
        {
            dbConnection.Exists<Bill>(o => Sql.In(o.Kvid, Kvids) && o.AmountPayment == o.Amount).ThrowIfTrue("此账单已归档付款登记,无法删除");
            if (dbConnection is LinqConnection && (dbConnection as LinqConnection).DbTransaction == null)
            {
                _trans = dbConnection.OpenTransaction();
            }
            return base.OnPreRestfulDelete(req, res, dbConnection, rtns);
        }

        public override bool OnPostRestfulDelete(IRequest req, IResponse res, IDbConnection dbConnection, IRestfulDeleteResponse<Bill> rtns)
        {
            try
            {
                var deleteBillDetails = dbConnection.From<BillDetail>().Update(o => o.Status).Where(o => Sql.In(o.BillKvid, Kvids));
                dbConnection.UpdateOnly<BillDetail>(new BillDetail { Status = -1 }, deleteBillDetails);

                _trans?.Commit();
                return base.OnPostRestfulDelete(req, res, dbConnection, rtns);
            }
            catch (Exception ex)
            {
                _trans?.Rollback();
                throw ex;
            }
            finally
            {
                _trans?.Dispose();
            }
        }
    }

    [Api(Description = "账单归档")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class BillArchiving : RestfulExecution<Bill>
    {
        public Guid Kvid { get; set; }

        public List<Guid> Kvids { get; set; }

        public override object OnExecution(IRequest req, IResponse res)
        {
            (Kvid == null && Kvids.IsNullOrEmpty()).ThrowIfTrue("请传入要结算的内容!");
            if (Kvids.IsNullOrEmpty()) Kvids = new List<Guid>();
            if (Kvid != null) Kvids.Add(Kvid);

            var conn = KiviiContext.GetOpenedDbConnection<Bill>();
            var bills = conn.SelectByIds<Bill>(Kvids);
            if (bills.IsNullOrEmpty()) throw new Exception("未找到传入的账单信息!");

            var rtns = new RestfulUpdateResponse<Bill>();
            rtns.Results = new List<Bill>();

            //var allBillDetails = conn.Select<BillDetail>(o => Sql.In(o.BillKvid, bills.ConvertAll(p => p.Kvid)));
            //var allSettlements = conn.Select<Settlement>(o => Sql.In(o.Kvid, allBillDetails.ConvertAll(p => p.BizKvid)));
            foreach (var bill in bills)
            {
                var item = bill.Archiving(conn);
                rtns.Results.Add(bill);
                //var billDetails = allBillDetails.Where(o => o.BillKvid == bill.Kvid).ToList();
                //if (billDetails.IsNullOrEmpty()) continue;
                //var settlements = allSettlements.Where(o => billDetails.ConvertAll(p => p.BizKvid).Contains(o.Kvid)).ToList();
                //if (settlements.IsNullOrEmpty())//未找到结算信息 说明此账单没有费用 
                //{
                //    bill.Amount = 0;
                //    bill.AddOnlyProperties(o => o.Amount);
                //    bill.AmountPlan = 0;
                //    bill.AddOnlyProperties(o => o.AmountPlan);
                //    bill.AmountPayment = 0;
                //    bill.AddOnlyProperties(o => o.AmountPayment);
                //    conn.UpdateOnly(bill);
                //    rtns.Results.Add(bill);
                //    continue;
                //}
                //bill.Amount = settlements.Sum(o => o.Amount);
                //bill.AddOnlyProperties(o => o.Amount);
                //bill.AmountPlan = settlements.Sum(o => o.AmountPlan);
                //bill.AddOnlyProperties(o => o.AmountPlan);
                //bill.AmountPayment = settlements.Sum(o => o.AmountPayment);
                //bill.AddOnlyProperties(o => o.AmountPayment);
                //conn.UpdateOnly(bill);
                //rtns.Results.Add(bill);
            }

            return rtns;
        }
    }
}
