﻿using Kivii.Finances.Entities;
using Kivii.Linq;
using Kivii.Web;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Text;
using System.Threading.Tasks;

namespace Kivii.Finances.Transforms
{
    public enum OwnerType
    {
        Unsupport = 0,
        Department = 1,
        Organization = 2,
        Member = 3
    }

    [Api(Description = "业务认领")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class PaymentSplit : RestfulExecution<Payment>
    {
        [ApiMember(Description = "认领金额")]
        public decimal Amount { get; set; }

        [ApiMember(Description = "认领类型,默认为Department,Organization,Member")]
        public OwnerType OwnerType { get; set; } = OwnerType.Department;

        public Guid OwnerKvid { get; set; }

        public string OwnerName { get; set; }

        public Guid Kvid { get; set; }

        public string Category { get; set; }

        public string Remark { get; set; }

        public override object OnExecution(IRequest req, IResponse res)
        {
            if (Kvid == Guid.Empty) throw new Exception("请选择要认领的收款!");

            var conn = KiviiContext.GetOpenedDbConnection<Payment>();

            var payment = conn.SingleById<Payment>(Kvid);
            if (payment == null) throw new Exception("收款不存在!");
            if (payment.Type != PaymentType.Bank && payment.Type != PaymentType.Split) throw new Exception("非银行到账不可以认领!");
            if (payment.OffsetKvid != Guid.Empty) throw new Exception("已冲销收款不可认领!");

            var ownerKvid = Guid.Empty;
            var ownerName = string.Empty;
            switch (OwnerType)
            {
                case OwnerType.Organization:
                    {
                        ownerKvid = KiviiContext.CurrentMember.OrganizationKvid;
                        ownerName = KiviiContext.CurrentMember.OrganizationName;
                        break;
                    }
                case OwnerType.Member:
                    {
                        ownerKvid = KiviiContext.CurrentMember.Kvid;
                        ownerName = KiviiContext.CurrentMember.FullName;
                        break;
                    }
                default:
                    {
                        ownerKvid = KiviiContext.CurrentMember.DepartmentKvid;
                        ownerName = KiviiContext.CurrentMember.DepartmentName;
                        break;
                    }
            }
            if (OwnerKvid != Guid.Empty)
            {
                ownerKvid = OwnerKvid;
                ownerName = OwnerName;
            }
            Payment bank = null;
            Payment split = null;
            EntityLog<Payment> log = null;

            if (payment.Type == PaymentType.Bank)
            {
                if (Amount <= 0) throw new Exception("金额不能为小于0!");
                if (payment.AmountSplited == payment.Amount) throw new Exception("无剩余收款可认领!");
                if (payment.AmountSplited + Amount > payment.Amount) throw new Exception("认领金额超出范围!");
                var amountSplit = conn.Scalar<Payment, decimal>(o => Sql.Sum(o.Amount), p => p.ParentKvid == payment.Kvid && p.OffsetKvid == Guid.Empty);
                if (amountSplit + Amount > payment.Amount) throw new Exception("认领金额超出范围!");

                bank = payment;
                split = bank.BizSplit(Amount, Remark, ownerKvid, ownerName, conn, Category);

                #region 记录日志
                log = new EntityLog<Payment>();
                log.OwnerKvid = bank.Kvid;
                log.BizId = split.SerialNumber;
                log.BizKvid = split.Kvid;
                log.BizType = typeof(Payment).FullName;
                log.Title = DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss");
                log.Remark = KiviiContext.CurrentMember.FullName + $"在  {log.Title} 对流水号为:{bank.SerialNumber}的收款进行认领至{ownerName},认领金额为:{split.Amount}元.";
                #endregion
            }

            if (payment.Type == PaymentType.Split)
            {
                split = payment;

                split.OwnerKvid = ownerKvid;
                split.AddOnlyProperties(o => o.OwnerKvid);
                split.OwnerName = ownerName;
                split.AddOnlyProperties(o => o.OwnerName);
                split.Remark = Remark;
                split.AddOnlyProperties(o => o.Remark);

                #region 记录日志
                log = new EntityLog<Payment>();
                log.OwnerKvid = split.ParentKvid;
                log.BizId = split.SerialNumber;
                log.BizKvid = split.Kvid;
                log.BizType = typeof(Payment).FullName;
                log.Title = DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss");
                log.Remark = KiviiContext.CurrentMember.FullName + $"在  {log.Title} 对流水号为:{split.SerialNumber}的收款进行认领至{ownerName},认领金额为:{split.Amount}元.";
                #endregion
            }
            var rtns = new RestfulExecutionResponse<Payment>();
            rtns.Results = new List<Payment>();

            //开启事务进行数据库操作
            var trans = conn.OpenTransaction();
            try
            {
                if (payment.Type == PaymentType.Bank)
                {
                    if (split != null)
                    {
                        conn.Insert(split);
                        rtns.Results.Add(split);
                    }
                    payment.RecalculateAmountSplit(conn);

                    if (log != null) conn.Insert(log);
                }
                if (payment.Type == PaymentType.Split)
                {
                    if (split != null)
                    {
                        conn.UpdateOnly(split);
                        rtns.Results.Add(split);
                    }
                    if (log != null) conn.Insert(log);
                }

                trans.Commit();
            }
            catch (Exception ex)
            {
                trans.Rollback();
                throw ex;
            }
            return rtns;
        }
    }

    [Api(Description = "批量业务认领")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class PaymentMultiSplit : RestfulExecution<Payment>
    {
        [ApiMember(Description = "认领类型,默认为Department,Organization,Member")]
        public OwnerType OwnerType { get; set; } = OwnerType.Department;

        public Guid OwnerKvid { get; set; }

        public string OwnerName { get; set; }

        public List<Guid> Kvids { get; set; }

        public string Category { get; set; }

        public string Remark { get; set; }
        public override object OnExecution(IRequest req, IResponse res)
        {
            if (Kvids.IsNullOrEmpty()) throw new Exception("请选择要认领的收款!");
            #region 启用缓存,将当前人创建批次数据存入,处理完毕或者接口报错后清除,确保不会重复创建
            var cache = KiviiContext.GetCacheClient();
            var cacheKey = KiviiContext.GetUrnKey($"{this.GetType().FullName}_Request_{KiviiContext.CurrentMember.FullName}_{KiviiContext.CurrentMember.Kvid}");
            var cacheValue = cache.Get<int?>(cacheKey);
            if (cacheValue != null)
            {
                if (cacheValue < 0) throw new Exception($"仍在处理中,请稍后查看结果!");
                else throw new Exception($"当前正在处理中,剩余{cacheValue}条,请稍后查看结果!");
            }
            int? caches = Kvids.Count;
            //caches.AddRange(Items.ConvertAll(o => o.SerialNumber));
            cache.Set(cacheKey, caches, TimeSpan.FromMinutes(5));
            #endregion

            try
            {
                var conn = KiviiContext.GetOpenedDbConnection<Payment>();

                var payments = conn.SelectByIds<Payment>(Kvids);
                if (payments == null) throw new Exception("收款不存在!");
                if (payments.Exists(o => o.Type != PaymentType.Bank && o.Type != PaymentType.Split)) throw new Exception("非银行到账不可以认领!");
                if (payments.Exists(o => o.OffsetKvid != Guid.Empty)) throw new Exception("已冲销收款不可认领!");
                if (payments.Exists(o => o.AmountSplited >= o.Amount)) throw new Exception("无剩余收款可认领!");

                var ownerKvid = Guid.Empty;
                var ownerName = string.Empty;
                switch (OwnerType)
                {
                    case OwnerType.Department:
                        {
                            ownerKvid = KiviiContext.CurrentMember.DepartmentKvid;
                            ownerName = KiviiContext.CurrentMember.DepartmentName;
                            break;
                        }
                    case OwnerType.Organization:
                        {
                            ownerKvid = KiviiContext.CurrentMember.OrganizationKvid;
                            ownerName = KiviiContext.CurrentMember.OrganizationName;
                            break;
                        }
                    case OwnerType.Member:
                        {
                            ownerKvid = KiviiContext.CurrentMember.Kvid;
                            ownerName = KiviiContext.CurrentMember.FullName;
                            break;
                        }
                }
                if (OwnerKvid != Guid.Empty)
                {
                    ownerKvid = OwnerKvid;
                    ownerName = OwnerName;
                }
                var banks = new List<Payment>();
                var splits = new List<Payment>();
                var splitsSetOwner = new List<Payment>();
                var logs = new List<EntityLog<Payment>>();

                foreach (var payment in payments)
                {
                    if (payment.Type == PaymentType.Bank)
                    {
                        var amount = payment.Amount - payment.AmountSplited;
                        if (amount <= 0) continue;
                        var splitPayment = payment.BizSplit(amount, Remark, ownerKvid, ownerName, conn, Category);
                        splits.Add(splitPayment);

                        banks.Add(payment);

                        #region 记录日志
                        var log = new EntityLog<Payment>();
                        log.OwnerKvid = payment.Kvid;
                        log.BizId = splitPayment.SerialNumber;
                        log.BizKvid = splitPayment.Kvid;
                        log.BizType = typeof(Payment).FullName;
                        log.Title = DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss");
                        log.Remark = KiviiContext.CurrentMember.FullName + $"在  {log.Title} 对流水号为:{payment.SerialNumber}的收款进行认领至{ownerName},认领金额为:{splitPayment.Amount}元.";
                        logs.Add(log);
                        #endregion
                    }
                    if (payment.Type == PaymentType.Split)
                    {
                        payment.OwnerKvid = ownerKvid;
                        payment.AddOnlyProperties(o => o.OwnerKvid);
                        payment.OwnerName = ownerName;
                        payment.AddOnlyProperties(o => o.OwnerName);
                        payment.Remark = Remark;
                        payment.AddOnlyProperties(o => o.Remark);
                        splitsSetOwner.Add(payment);

                        #region 记录日志
                        var log = new EntityLog<Payment>();
                        log.OwnerKvid = payment.ParentKvid;
                        log.BizId = payment.SerialNumber;
                        log.BizKvid = payment.Kvid;
                        log.BizType = typeof(Payment).FullName;
                        log.Title = DateTime.Now.ToString("yyyy-MM-dd HH:mm:ss");
                        log.Remark = KiviiContext.CurrentMember.FullName + $"在  {log.Title} 对流水号为:{payment.SerialNumber}的收款进行认领至{ownerName},认领金额为:{payment.Amount}元.";
                        logs.Add(log);
                        #endregion
                    }
                }

                var rtns = new RestfulExecutionResponse<Payment>();
                rtns.Results = new List<Payment>();


                //开启事务进行数据库操作
                var trans = conn.OpenTransaction();
                try
                {
                    foreach (var split in splits)
                    {
                        conn.Insert(split);
                        rtns.Results.Add(split);
                        caches--;
                        cache.Replace(cacheKey, caches);
                    }
                    foreach (var item in splitsSetOwner)
                    {
                        conn.UpdateOnly(item);
                        rtns.Results.Add(item);
                    }
                    foreach (var item in banks)
                    {
                        item.RecalculateAmountSplit(conn);
                    }
                    foreach (var log in logs)
                    {
                        conn.Insert(log);
                    }

                    trans.Commit();
                    cache.Remove(cacheKey);
                }
                catch (Exception ex)
                {
                    trans.Rollback();
                    cache.Remove(cacheKey);
                    throw ex;
                }
                return rtns;
            }
            catch (Exception ex)
            {
                cache.Remove(cacheKey);
                throw ex;
            }
        }
    }

    [Api(Description = "认领撤销")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class PaymentSplitCancel : RestfulExecution<Payment>
    {
        public List<Guid> Kvids { get; set; }

        public string Remark { get; set; }

        public override object OnExecution(IRequest req, IResponse res)
        {
            if (Kvids.IsNullOrEmpty()) throw new Exception("请选择要撤销认领的收款!");

            var conn = KiviiContext.GetOpenedDbConnection<Payment>();
            var payments = conn.SelectByIds<Payment>(Kvids);
            if (payments == null) throw new Exception("收款不存在!");
            if (payments.Exists(o => o.Type != PaymentType.Split)) throw new Exception("非银行到账不可以认领!");
            if (payments.Exists(o => o.OffsetKvid != Guid.Empty)) throw new Exception("已冲销收款不可撤销认领!");
            (conn.Exists<InvoiceApply>(o => o.OperateType == InvoiceApplyType.Related && o.OffsetKvid == Guid.Empty && o.AmountInvoice != o.Amount && Sql.In(o.BizKvid, payments.ConvertAll(p => p.Kvid)))).ThrowIfTrue("存在申请开票中的到账,暂无法撤销认领!");
            (conn.Exists<Payment>(o => o.OffsetKvid == Guid.Empty && o.Type == PaymentType.Assign && Sql.In(o.ParentKvid, payments.ConvertAll(p => p.Kvid)))).ThrowIfTrue("认领金额存在已经划转,请先撤销划转以继续!");
            var rtns = new RestfulExecutionResponse<Payment>();
            rtns.Results = new List<Payment>();
            //if (payments.Exists(o => o.AmountUsed > 0)) throw new Exception("存在到账已使用,不可撤销认领!");
            //认领撤销分两种情况.
            //第一种:要是要撤销的目标已经关联过业务了被使用了,撤销只会清空所属信息,重新认领则只是赋值所属信息
            //第二种:要是没有被业务关联没被使用,撤销则是Offset当前拆分项目,重新认领则会生成新的拆分项

            var paymentUnuseds = payments.Where(o => o.AmountUsed == 0).ToList();
            var paymentUseds = payments.Where(o => o.AmountUsed > 0).ToList();

            #region 已经使用了的撤销
            if (!paymentUseds.IsNullOrEmpty())
            {
                var paymentKvids = paymentUseds.ConvertAll(o => o.Kvid);

                var updateLamda = conn.From<Payment>();
                updateLamda = updateLamda.Update(o => new { o.OwnerKvid, o.OwnerName });
                updateLamda = updateLamda.Where(o => Sql.In(o.Kvid, paymentKvids));
                conn.UpdateOnly(new Payment { OwnerKvid = Guid.Empty, OwnerName = string.Empty }, updateLamda);
                rtns.Results.AddRange(paymentUseds);
            }
            #endregion

            #region 没有使用的撤销
            if (!paymentUnuseds.IsNullOrEmpty())
            {
                var parentKvids = paymentUnuseds.ConvertAll(o => o.ParentKvid).Distinct().ToList();
                var parents = conn.SelectByIds<Payment>(parentKvids);

                var parentUpdates = new List<Payment>();
                var splitOffsets = new List<Payment>();
                var offsets = new List<Payment>();

                foreach (var split in paymentUnuseds)
                {
                    var parent = parents.FirstOrDefault(o => o.Kvid == split.ParentKvid);
                    if (parent == null) continue;

                    var offset = split.Offset($"认领撤销:{Remark}");
                    offsets.Add(offset);
                    splitOffsets.Add(split);
                    parentUpdates.Add(parent);
                }

                var trans = conn.OpenTransaction();
                try
                {
                    foreach (var item in splitOffsets)
                    {
                        conn.UpdateOnly(item);
                        rtns.Results.Add(item);
                    }
                    foreach (var item in offsets)
                    {
                        conn.Insert(item);
                    }
                    foreach (var item in parentUpdates)
                    {
                        item.RecalculateAmountSplit(conn);
                    }
                    trans.Commit();
                }
                catch (Exception ex)
                {
                    trans.Rollback();
                    throw ex;
                }
            }

            #endregion

            return rtns;
        }
    }

    [Api(Description = "到账所属")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class PaymentSetOwner : RestfulExecution<Payment>
    {
        public OwnerType OwnerType { get; set; } = OwnerType.Department;

        public Guid OwnerKvid { get; set; }

        public string OwnerName { get; set; }

        public List<Guid> Kvids { get; set; }

        public string Remark { get; set; }

        public override object OnExecution(IRequest req, IResponse res)
        {
            if (Kvids.IsNullOrEmpty()) throw new Exception("请选择要标记所属部门的收款!");
            var conn = KiviiContext.GetOpenedDbConnection<Payment>();
            var payments = conn.SelectByIds<Payment>(Kvids);
            if (payments == null) throw new Exception("收款不存在!");
            if (payments.Exists(o => o.Type != PaymentType.Bank)) throw new Exception("非银行到账不可继续操作!");
            if (payments.Exists(o => o.OffsetKvid != Guid.Empty)) throw new Exception("已作废收款不可继续操作!");
            var paymentKvids = payments.ConvertAll(o => o.Kvid);
            if (OwnerKvid == Guid.Empty)
            {
                switch (OwnerType)
                {
                    case OwnerType.Department:
                        {
                            OwnerKvid = KiviiContext.CurrentMember.DepartmentKvid;
                            OwnerName = KiviiContext.CurrentMember.DepartmentName;
                            break;
                        }
                    case OwnerType.Organization:
                        {
                            OwnerKvid = KiviiContext.CurrentMember.OrganizationKvid;
                            OwnerName = KiviiContext.CurrentMember.OrganizationName;
                            break;
                        }
                    case OwnerType.Member:
                        {
                            OwnerKvid = KiviiContext.CurrentMember.Kvid;
                            OwnerName = KiviiContext.CurrentMember.FullName;
                            break;
                        }
                }
            }
            var rtns = new RestfulExecutionResponse<Payment>();
            rtns.Results = new List<Payment>();

            var updateLamda = conn.From<Payment>();
            updateLamda = updateLamda.Update(o => new { o.OwnerKvid, o.OwnerName, o.BizId });
            updateLamda = updateLamda.Where(o => Sql.In(o.Kvid, paymentKvids));
            conn.UpdateOnly(new Payment { OwnerKvid = OwnerKvid, OwnerName = OwnerName, BizId = Remark }, updateLamda);
            foreach (var item in payments)
            {
                item.OwnerName = OwnerName;
                item.OwnerKvid = OwnerKvid;
                item.BizId = Remark;
            }
            rtns.Results.AddRange(payments);
            return rtns;
        }
    }

    [Api(Description = "清除到账所属")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class PaymentCleanOwner : RestfulExecution<Payment>
    {
        public List<Guid> Kvids { get; set; }

        public override object OnExecution(IRequest req, IResponse res)
        {
            if (Kvids.IsNullOrEmpty()) throw new Exception("请选择要标记所属部门的收款!");
            var conn = KiviiContext.GetOpenedDbConnection<Payment>();
            var payments = conn.SelectByIds<Payment>(Kvids);
            if (payments == null) throw new Exception("收款不存在!");
            if (payments.Exists(o => o.Type != PaymentType.Bank)) throw new Exception("非银行到账不可继续操作!");
            if (payments.Exists(o => o.OffsetKvid != Guid.Empty)) throw new Exception("已作废收款不可继续操作!");
            var paymentKvids = payments.ConvertAll(o => o.Kvid);
            var rtns = new RestfulExecutionResponse<Payment>();
            rtns.Results = new List<Payment>();

            var updateLamda = conn.From<Payment>();
            updateLamda = updateLamda.Update(o => new { o.OwnerKvid, o.OwnerName });
            updateLamda = updateLamda.Where(o => Sql.In(o.Kvid, paymentKvids));
            conn.UpdateOnly(new Payment { OwnerKvid = Guid.Empty, OwnerName = string.Empty }, updateLamda);
            foreach (var item in payments)
            {
                item.OwnerName = string.Empty;
                item.OwnerKvid = Guid.Empty;
            }
            rtns.Results.AddRange(payments);
            return rtns;
        }
    }

    [Api(Description = "认领划转")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class PaymentAssign : RestfulExecution<Payment>
    {
        [ApiMember(Description = "划转金额")]
        public decimal Amount { get; set; }

        [ApiMember(Description = "所属类型,默认为Department,Organization,Member")]
        public OwnerType OwnerType { get; set; } = OwnerType.Department;

        public bool AmountFirst { get; set; } = true;//标记划转出去的金额没有开票和没有使用优先,默认true,优先划转已经开票和使用的金额
        public Guid OwnerKvid { get; set; }

        public string OwnerName { get; set; }

        public Guid Kvid { get; set; }

        public string Remark { get; set; }

        public override object OnExecution(IRequest req, IResponse res)
        {
            if (Kvid == Guid.Empty) throw new Exception("请选择要划转的收款!");

            var conn = KiviiContext.GetOpenedDbConnection<Payment>();

            var payment = conn.SingleById<Payment>(Kvid);
            if (payment == null) throw new Exception("收款不存在!");
            if (payment.Type != PaymentType.Split) throw new Exception("请选择正确类型!");
            if (payment.OffsetKvid != Guid.Empty) throw new Exception("已冲销收款不可划转!");
            if (Amount <= 0) throw new Exception("金额不能为小于0!");
            if (payment.AmountSplited == payment.Amount) throw new Exception("无剩余收款可划转!");
            if (payment.AmountSplited + Amount > payment.Amount) throw new Exception("划转金额超出范围!");
            var amountAssign = conn.Scalar<Payment, decimal>(o => Sql.Sum(o.Amount), p => p.ParentKvid == payment.Kvid && p.Type == PaymentType.Assign && p.OffsetKvid == Guid.Empty);
            if (amountAssign + Amount > payment.Amount) throw new Exception("划转金额超出范围!");
            var ownerKvid = Guid.Empty;
            var ownerName = string.Empty;
            switch (OwnerType)
            {
                case OwnerType.Organization:
                    {
                        ownerKvid = KiviiContext.CurrentMember.OrganizationKvid;
                        ownerName = KiviiContext.CurrentMember.OrganizationName;
                        break;
                    }
                case OwnerType.Member:
                    {
                        ownerKvid = KiviiContext.CurrentMember.Kvid;
                        ownerName = KiviiContext.CurrentMember.FullName;
                        break;
                    }
                default:
                    {
                        ownerKvid = KiviiContext.CurrentMember.DepartmentKvid;
                        ownerName = KiviiContext.CurrentMember.DepartmentName;
                        break;
                    }
            }
            if (OwnerKvid != Guid.Empty)
            {
                ownerKvid = OwnerKvid;
                ownerName = OwnerName;
            }
            if (payment.OwnerKvid == ownerKvid) throw new Exception("不可划转至同部门!");
            var assign = payment.BizAssign(Amount, Remark, ownerKvid, ownerName, conn, AmountFirst);

            var rtns = new RestfulExecutionResponse<Payment>();
            rtns.Results = new List<Payment>();

            //开启事务进行数据库操作
            var trans = conn.OpenTransaction();
            try
            {
                if (assign != null)
                {
                    conn.Insert(assign);
                    rtns.Results.Add(assign);
                }
                payment.RecalculateAmountSplit(conn);

                trans.Commit();
            }
            catch (Exception ex)
            {
                trans.Rollback();
                throw ex;
            }
            return rtns;
        }
    }

    [Api(Description = "认领批量划转")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class PaymentMultiAssign : RestfulExecution<Payment>
    {
        [ApiMember(Description = "所属类型,默认为Department,Organization,Member")]
        public OwnerType OwnerType { get; set; } = OwnerType.Department;

        public Guid OwnerKvid { get; set; }

        public string OwnerName { get; set; }
        public bool AmountFirst { get; set; } = true;//标记划转出去的金额没有开票和没有使用优先,默认true,优先划转已经开票和使用的金额

        public List<Guid> Kvids { get; set; }

        public string Remark { get; set; }

        public override object OnExecution(IRequest req, IResponse res)
        {
            if (Kvids.IsNullOrEmpty()) throw new Exception("请选择要划转的收款!");

            #region 启用缓存,将当前人创建批次数据存入,处理完毕或者接口报错后清除,确保不会重复创建
            var cache = KiviiContext.GetCacheClient();
            var cacheKey = KiviiContext.GetUrnKey($"{this.GetType().FullName}_Request_{KiviiContext.CurrentMember.FullName}_{KiviiContext.CurrentMember.Kvid}");
            var cacheValue = cache.Get<int?>(cacheKey);
            if (cacheValue != null)
            {
                if (cacheValue < 0) throw new Exception($"仍在处理中,请稍后查看结果!");
                else throw new Exception($"当前正在处理中,剩余{cacheValue}条,请稍后查看结果!");
            }
            int? caches = Kvids.Count;
            //caches.AddRange(Items.ConvertAll(o => o.SerialNumber));
            cache.Set(cacheKey, caches, TimeSpan.FromMinutes(5));
            #endregion

            try
            {
                var conn = KiviiContext.GetOpenedDbConnection<Payment>();

                var payments = conn.SelectByIds<Payment>(Kvids);
                if (payments == null) throw new Exception("收款不存在!");
                if (payments.Exists(o => o.Type != PaymentType.Split)) throw new Exception("请选择正确类型!");
                if (payments.Exists(o => o.OffsetKvid != Guid.Empty)) throw new Exception("已冲销收款不可划转!");
                if (payments.Exists(o => o.AmountSplited >= o.Amount)) throw new Exception("无剩余收款可划转!");

                var ownerKvid = Guid.Empty;
                var ownerName = string.Empty;
                switch (OwnerType)
                {
                    case OwnerType.Department:
                        {
                            ownerKvid = KiviiContext.CurrentMember.DepartmentKvid;
                            ownerName = KiviiContext.CurrentMember.DepartmentName;
                            break;
                        }
                    case OwnerType.Organization:
                        {
                            ownerKvid = KiviiContext.CurrentMember.OrganizationKvid;
                            ownerName = KiviiContext.CurrentMember.OrganizationName;
                            break;
                        }
                    case OwnerType.Member:
                        {
                            ownerKvid = KiviiContext.CurrentMember.Kvid;
                            ownerName = KiviiContext.CurrentMember.FullName;
                            break;
                        }
                }
                if (OwnerKvid != Guid.Empty)
                {
                    ownerKvid = OwnerKvid;
                    ownerName = OwnerName;
                }
                if (payments.Exists(o => o.OwnerKvid == ownerKvid)) throw new Exception("不可划转至同部门!");

                var spliteds = new List<Payment>();
                var assigns = new List<Payment>();

                foreach (var payment in payments)
                {
                    var amount = payment.Amount - payment.AmountSplited;
                    if (amount <= 0) continue;
                    var assignPayment = payment.BizAssign(amount, Remark, ownerKvid, ownerName, conn, AmountFirst);
                    assigns.Add(assignPayment);

                    spliteds.Add(payment);
                }

                var rtns = new RestfulExecutionResponse<Payment>();
                rtns.Results = new List<Payment>();

                //开启事务进行数据库操作
                var trans = conn.OpenTransaction();
                try
                {
                    foreach (var assign in assigns)
                    {
                        conn.Insert(assign);
                        rtns.Results.Add(assign);
                        caches--;
                        cache.Replace(cacheKey, caches);
                    }
                    foreach (var item in spliteds)
                    {
                        item.RecalculateAmountSplit(conn);
                    }
                    trans.Commit();
                    cache.Remove(cacheKey);
                }
                catch (Exception ex)
                {
                    trans.Rollback();
                    cache.Remove(cacheKey);
                    throw ex;
                }
                return rtns;
            }
            catch (Exception ex)
            {
                cache.Remove(cacheKey);
                throw ex;
            }
        }
    }

    [Api(Description = "划转撤销")]
    [RequiresAnyRole(SystemRoles.Everyone)]
    public class PaymentAssignCancel : RestfulExecution<Payment>
    {
        public List<Guid> Kvids { get; set; }

        public string Remark { get; set; }

        public override object OnExecution(IRequest req, IResponse res)
        {
            if (Kvids.IsNullOrEmpty()) throw new Exception("请选择要撤销划转的收款!");

            var conn = KiviiContext.GetOpenedDbConnection<Payment>();
            var payments = conn.SelectByIds<Payment>(Kvids);
            if (payments == null) throw new Exception("收款不存在!");
            if (payments.Exists(o => o.Type != PaymentType.Assign)) throw new Exception("请选择正确类型!");
            if (payments.Exists(o => o.OffsetKvid != Guid.Empty)) throw new Exception("已冲销收款不可撤销划转!");
            (conn.Exists<InvoiceApply>(o => o.OperateType == InvoiceApplyType.Related && o.OffsetKvid == Guid.Empty && o.AmountInvoice != o.Amount && Sql.In(o.BizKvid, payments.ConvertAll(p => p.Kvid)))).ThrowIfTrue("存在申请开票中的到账,暂无法撤销认领!");
            conn.Exists<Payment>(o => o.OffsetKvid == Guid.Empty && o.Amount > 0 && Sql.In(o.Type, PaymentType.Transfer, PaymentType.DynamicPay) && Sql.In(o.ParentKvid, payments.ConvertAll(p => p.Kvid))).ThrowIfTrue("存在已收款登记或账户充值的划转!");
            var rtns = new RestfulExecutionResponse<Payment>();
            rtns.Results = new List<Payment>();

            var parentKvids = payments.ConvertAll(o => o.ParentKvid).Distinct().ToList();
            var parentSplits = conn.SelectByIds<Payment>(parentKvids);

            var parentUpdates = new List<Payment>();
            var assignOffsets = new List<Payment>();
            var offsets = new List<Payment>();

            foreach (var assignPayment in payments)
            {
                var parentSplit = parentSplits.FirstOrDefault(o => o.Kvid == assignPayment.ParentKvid);
                if (parentSplit == null) continue;

                var offset = assignPayment.Offset($"划转撤销:{Remark}");
                offsets.Add(offset);
                assignOffsets.Add(assignPayment);
                parentUpdates.Add(parentSplit);
            }

            var trans = conn.OpenTransaction();
            try
            {
                foreach (var item in assignOffsets)
                {
                    conn.UpdateOnly(item);
                    rtns.Results.Add(item);
                }
                foreach (var item in offsets)
                {
                    conn.Insert(item);
                }
                foreach (var item in parentUpdates)
                {
                    item.RecalculateAmountSplit(conn);
                }
                trans.Commit();
            }
            catch (Exception ex)
            {
                trans.Rollback();
                throw ex;
            }

            return rtns;
        }
    }
}
